X(DP8KC_REGMODE_A_NOREG_REGMODE_B_NOREG)
X(DP8KC_REGMODE_A_NOREG_REGMODE_B_OUTREG)
X(DP8KC_REGMODE_A_OUTREG_REGMODE_B_NOREG)
X(DP8KC_REGMODE_A_OUTREG_REGMODE_B_OUTREG)
X(DP8KC_WRITEMODE_A_NORMAL_WRITEMODE_B_NORMAL)
X(DP8KC_WRITEMODE_A_NORMAL_WRITEMODE_B_READBEFOREWRITE)
X(DP8KC_WRITEMODE_A_NORMAL_WRITEMODE_B_WRITETHROUGH)
X(PIO_IOTYPE_LVCMOS12)
X(PIO_IOTYPE_LVCMOS15)
X(PIO_IOTYPE_LVCMOS18)
X(PIO_IOTYPE_LVCMOS25)
X(PIO_IOTYPE_LVCMOS33)
X(PIO_IOTYPE_LVDS)
X(PIO_IOTYPE_SSTL15_I)
X(PIO_IOTYPE_SSTL15_II)
X(PIO_IOTYPE_SSTL18_I)
X(PIO_IOTYPE_SSTL18_II)
X(PIO_IOTYPE_LVDS25)
X(PIO_IOTYPE_LVPECL33)
X(PIO_IOTYPE_LVTTL33)
X(PIO_IOTYPE_MIPI)
X(PAD)
X(PADDI)
X(PADDO)
X(PADDT)

X(TRELLIS_SLICE)
X(A0)
X(B0)
X(C0)
X(D0)
X(A1)
X(B1)
X(C1)
X(D1)
X(M0)
X(M1)
X(FCI)
X(FXA)
X(FXB)
X(CLK)
X(LSR)
X(CE)
X(CEA)
X(CEW)
X(DI0)
X(DI1)
X(WD0)
X(WD1)
X(WAD0)
X(WAD1)
X(WAD2)
X(WAD3)
X(WRE)
X(WCK)
X(F0)
X(Q0)
X(F1)
X(Q1)
X(FCO)
X(OFX0)
X(OFX1)
X(WDO0)
X(WDO1)
X(WDO2)
X(WDO3)
X(WADO0)
X(WADO1)
X(WADO2)
X(WADO3)

X(MODE)
X(GSR)
X(SGSR)
X(SRMODE)
X(CEMUX)
X(CLKMUX)
X(LSRMUX)
X(LSRONMUX)
X(LUT0_INITVAL)
X(LUT1_INITVAL)
X(REGMODE)
X(REG0_SD)
X(REG1_SD)
X(REG0_REGSET)
X(REG1_REGSET)
X(CCU2_INJECT1_0)
X(CCU2_INJECT1_1)
X(WREMUX)


X(TRELLIS_FF)
X(DI)
X(Q)

X(REGSET)


X(TRELLIS_IO)
X(I)
X(EN)
X(O)

X(DIR)


X(LUT4)
X(A)
X(B)
X(C)
X(D)
X(Z)

X(INIT)


X(PFUMX)
X(ALUT)
X(BLUT)


X(L6MUX21)
X(SD)


X(T)
X(IOLDO)
X(IOLTO)
X(PG)
X(INRD)
X(LVDS)

X(BEL)
X(GND)
X(IO_TYPE)
X(LOC)
X(NOM_FREQ)
X(VCC)

X(WIRE_TYPE_NONE)
X(TILE_WIRE_ID)

X(machxo2)
X(pack)
X(place)
X(placer)
X(route)
X(router)


X(TRELLIS_COMB)
X(TRELLIS_RAMW)
X(WD)
X(OFX)
X(F)
X(M)

X(CCU2D)
X(CCU2_INJECT1)
X(INJECT1_0)
X(INJECT1_1)
X(TRELLIS_DPR16X4)
X(INITVAL)
X(INIT0)
X(INIT1)

X(DATAMUX_ODDR)
X(DATAMUX_MDDR)
X(CIN)
X(S0)
X(S1)
X(COUT)

X(E)
X(Y)

X(WCKMUX)

X(EHXPLLJ)
X(CLKFB)
X(PHASESEL0)
X(PHASESEL1)
X(PHASEDIR)
X(PHASESTEP)
X(LOADREG)
X(STDBY)
X(PLLWAKESYNC)
X(RST)
X(RESETM)
X(RESETC)
X(RESETD)
X(ENCLKOP)
X(ENCLKOS)
X(ENCLKOS2)
X(ENCLKOS3)
X(PLLCLK)
X(PLLRST)
X(PLLSTB)
X(PLLWE)
X(PLLDATI0)
X(PLLDATI1)
X(PLLDATI2)
X(PLLDATI3)
X(PLLDATI4)
X(PLLDATI5)
X(PLLDATI6)
X(PLLDATI7)
X(PLLADDR0)
X(PLLADDR1)
X(PLLADDR2)
X(PLLADDR3)
X(PLLADDR4)
X(CLKOP)
X(CLKOS)
X(CLKOS2)
X(CLKOS3)
X(LOCK)
X(INTLOCK)
X(REFCLK)
X(CLKINTFB)
X(DPHSRC)
X(PLLDATO0)
X(PLLDATO1)
X(PLLDATO2)
X(PLLDATO3)
X(PLLDATO4)
X(PLLDATO5)
X(PLLDATO6)
X(PLLDATO7)
X(PLLACK)

X(DP8KC)
X(PDPW8KC)
X(FIFO8KB)

X(DIA0)
X(DIA1)
X(DIA2)
X(DIA3)
X(DIA4)
X(DIA5)
X(DIA6)
X(DIA7)
X(DIA8)
X(ADA0)
X(ADA1)
X(ADA2)
X(ADA3)
X(ADA4)
X(ADA5)
X(ADA6)
X(ADA7)
X(ADA8)
X(ADA9)
X(ADA10)
X(ADA11)
X(ADA12)
X(OCEA)
X(CLKA)
X(WEA)
X(CSA2)
X(CSA1)
X(CSA0)
X(RSTA)
X(DIB0)
X(DIB1)
X(DIB2)
X(DIB3)
X(DIB4)
X(DIB5)
X(DIB6)
X(DIB7)
X(DIB8)
X(ADB0)
X(ADB1)
X(ADB2)
X(ADB3)
X(ADB4)
X(ADB5)
X(ADB6)
X(ADB7)
X(ADB8)
X(ADB9)
X(ADB10)
X(ADB11)
X(ADB12)
X(CEB)
X(OCEB)
X(CLKB)
X(WEB)
X(CSB2)
X(CSB1)
X(CSB0)
X(RSTB)
X(DOA0)
X(DOA1)
X(DOA2)
X(DOA3)
X(DOA4)
X(DOA5)
X(DOA6)
X(DOA7)
X(DOA8)
X(DOB0)
X(DOB1)
X(DOB2)
X(DOB3)
X(DOB4)
X(DOB5)
X(DOB6)
X(DOB7)
X(DOB8)
X(AE)
X(AF)
X(EF)
X(FF)
X(WID)
X(CLKAMUX)
X(CLKBMUX)
X(CEAMUX)
X(CEBMUX)
X(OCEAMUX)
X(OCEBMUX)

X(CLKI_DIV)
X(CLKFB_DIV)
X(CLKOP_DIV)
X(CLKOS_DIV)
X(CLKOS2_DIV)
X(CLKOS3_DIV)
X(CLOCK_ENABLE_PORTS)
X(CLKOP_ENABLE)
X(CLKOS_ENABLE)
X(CLKOS2_ENABLE)
X(CLKOS3_ENABLE)
X(VCO_BYPASS_A0)
X(VCO_BYPASS_B0)
X(VCO_BYPASS_C0)
X(VCO_BYPASS_D0)
X(CLKOP_CPHASE)
X(CLKOS_CPHASE)
X(CLKOS2_CPHASE)
X(CLKOS3_CPHASE)
X(CLKOP_FPHASE)
X(CLKOS_FPHASE)
X(CLKOS2_FPHASE)
X(CLKOS3_FPHASE)
X(FEEDBK_PATH)
X(KVCO)
X(LPF_CAPACITOR)
X(LPF_RESISTOR)
X(ICP_CURRENT)
X(FRACN_ENABLE)
X(FRACN_DIV)
X(FRACN_ORDER)
X(CLKOP_TRIM_POL)
X(CLKOP_TRIM_DELAY)
X(CLKOS_TRIM_POL)
X(CLKOS_TRIM_DELAY)
X(PLL_EXPERT)
X(PLL_USE_WB)
X(PREDIVIDER_MUXA1)
X(PREDIVIDER_MUXB1)
X(PREDIVIDER_MUXC1)
X(PREDIVIDER_MUXD1)
X(OUTDIVIDER_MUXA2)
X(OUTDIVIDER_MUXB2)
X(OUTDIVIDER_MUXC2)
X(OUTDIVIDER_MUXD2)
X(FREQ_LOCK_ACCURACY)
X(PLL_LOCK_MODE)
X(PLL_LOCK_DELAY)
X(GMC_GAIN)
X(GMC_TEST)
X(MFG1_TEST)
X(MFG2_TEST)
X(MFG_FORCE_VFILTER)
X(MFG_ICP_TEST)
X(MFG_EN_UP)
X(MFG_FLOAT_ICP)
X(MFG_GMC_PRESET)
X(MFG_LF_PRESET)
X(MFG_GMC_RESET)
X(MFG_LF_RESET)
X(MFG_LF_RESGRND)
X(MFG_GMCREF_SEL)
X(MFG_ENABLE_FILTEROPAMP)
X(STDBY_ENABLE)
X(REFIN_RESET)
X(SYNC_ENABLE)
X(INT_LOCK_STICKY)
X(DPHASE_SOURCE)
X(INTFB_WAKE)
X(PLLRST_ENA)
X(MRST_ENA)
X(DCRST_ENA)
X(DDRST_ENA)

X(DATA_WIDTH_A)
X(DATA_WIDTH_B)
X(WRITEMODE_A)
X(WRITEMODE_B)
X(CSDECODE_A)
X(CSDECODE_B)
X(REGMODE_A)
X(REGMODE_B)
X(RESETMODE)
X(ASYNC_RESET_RELEASE)
X(RSTAMUX)
X(RSTBMUX)
X(WEAMUX)
X(WEBMUX)
X(DATA_WIDTH_W)
X(DATA_WIDTH_R)
X(CSDECODE_W)
X(CSDECODE_R)
X(CSDECODE)
X(AEPOINTER)
X(AEPOINTER1)
X(AFPOINTER)
X(AFPOINTER1)
X(FULLPOINTER)
X(FULLPOINTER1)
X(CLKW)
X(CLKR)
X(CER)
X(OCER)

X(OSCH)
X(OSC)
X(SEDSTDBY)

X(DCCA)
X(CLKI)
X(CLKO)

X(DCMA)
X(CLK0)
X(CLK1)
X(SEL)
X(DCMOUT)

X(EFB)
X(WBCUFMIRQ)
X(WBACKO)
X(WBDATO0)
X(WBDATO1)
X(WBDATO2)
X(WBDATO3)
X(WBDATO4)
X(WBDATO5)
X(WBDATO6)
X(WBDATO7)
X(WBDATI0)
X(WBDATI1)
X(WBDATI2)
X(WBDATI3)
X(WBDATI4)
X(WBDATI5)
X(WBDATI6)
X(WBDATI7)
X(WBADRI0)
X(WBADRI1)
X(WBADRI2)
X(WBADRI3)
X(WBADRI4)
X(WBADRI5)
X(WBADRI6)
X(WBADRI7)
X(WBWEI)
X(WBSTBI)
X(WBCYCI)
X(WBRSTI)
X(WBCLKI)
X(CFGWAKE)
X(CFGSTDBY)
X(UFMSN)
X(PLL0ACKI)
X(PLL0STBO)
X(PLL0DATI0)
X(PLL0DATI1)
X(PLL0DATI2)
X(PLL0DATI3)
X(PLL0DATI4)
X(PLL0DATI5)
X(PLL0DATI6)
X(PLL0DATI7)
X(PLL1ACKI)
X(PLL1STBO)
X(PLL1DATI0)
X(PLL1DATI1)
X(PLL1DATI2)
X(PLL1DATI3)
X(PLL1DATI4)
X(PLL1DATI5)
X(PLL1DATI6)
X(PLL1DATI7)
X(PLLADRO0)
X(PLLADRO1)
X(PLLADRO2)
X(PLLADRO3)
X(PLLADRO4)
X(PLLWEO)
X(PLLRSTO)
X(PLLCLKO)
X(TCOC)
X(TCINT)
X(TCIC)
X(TCRSTN)
X(TCCLKI)
X(SPIIRQO)
X(SPISCSN)
X(SPICSNEN)
X(SPIMOSIEN)
X(SPIMOSIO)
X(SPIMOSII)
X(SPIMISOEN)
X(SPIMISOO)
X(SPIMISOI)
X(SPISCKEN)
X(SPISCKO)
X(SPISCKI)
X(SPIMCSN0)
X(SPIMCSN1)
X(SPIMCSN2)
X(SPIMCSN3)
X(SPIMCSN4)
X(SPIMCSN5)
X(SPIMCSN6)
X(SPIMCSN7)
X(I2C1IRQO)
X(I2C1SDAOEN)
X(I2C1SDAO)
X(I2C1SDAI)
X(I2C1SCLOEN)
X(I2C1SCLO)
X(I2C1SCLI)
X(I2C2IRQO)
X(I2C2SDAOEN)
X(I2C2SDAO)
X(I2C2SDAI)
X(I2C2SCLOEN)
X(I2C2SCLO)
X(I2C2SCLI)

X(JTAGF)
X(TCK)
X(TMS)
X(TDI)
X(JTDO2)
X(JTDO1)
X(TDO)
X(JTDI)
X(JTCK)
X(JRTI2)
X(JRTI1)
X(JSHIFT)
X(JUPDATE)
X(JRSTN)
X(JCE2)
X(JCE1)
X(ER1)
X(ER2)

X(PCNTR)
X(CLRFLAG)
X(USERSTDBY)
X(USERTIMEOUT)
X(SFLAG)
X(STOP)

X(SEDFA)
X(SEDENABLE)
X(SEDSTART)
X(SEDFRCERR)
X(SEDEXCLK)
X(SEDERR)
X(SEDDONE)
X(SEDINPROG)
X(AUTODONE)

X(SYNCMODE)

X(START)
X(STARTCLK)

X(TSALL)
X(TSALLI)

X(CLKDIVC)
X(ALIGNWD)
X(CDIV1)
X(CDIVX)

X(ECLKSYNCA)
X(ECLKI)
X(ECLKO)

X(ECLKBRIDGECS)
X(ECSOUT)


X(RST_PULSE)
X(DIV)
X(SED_CLK_FREQ)
X(CHECKALWAYS)
X(DEV_DENSITY)
X(DEL_ADJ)
X(DEL_VAL)
X(LOCK_SENSITIVITY)
X(FIN)
X(FORCE_MAX_DELAY)
X(DEL_MODE)
X(DEL_VALUE)
X(STDBYOPT)
X(TIMEOUT)
X(WAKEUP)
X(POROFF)
X(BGOFF)
X(BANKID)

X(DLLDELC)
X(DQSDEL)

X(PLLCSOUT)
X(PLLREFCS)

X(CLKFBBUF)

X(DQSDLLC)
X(UDDCNTLN)
X(FREEZE)

X(IOLOGIC)
X(BIOLOGIC)
X(BSIOLOGIC)
X(TIOLOGIC)
X(TSIOLOGIC)
X(RIOLOGIC)

X(INDD)
X(TDX0)
X(TDX1)
X(TDX2)
X(TDX3)
X(TDX4)
X(TDX5)
X(TDX6)
X(TDX7)
X(DEL0)
X(DEL1)
X(DEL2)
X(DEL3)
X(DEL4)
X(ONEG)
X(OPOS)
X(TS)
X(ECLK)
X(SLIP)
X(DDRCLKPOL)
X(DQSR90)
X(DQSW90)
X(IN)
X(IP)
X(RXD0)
X(RXD1)
X(RXD2)
X(RXD3)
X(RXDA0)
X(RXDA1)
X(RXDA2)
X(RXDA3)
X(RXDA4)
X(RXDA5)
X(RXDA6)
X(RXDA7)

X(OSCJ)
X(OSCESB)

X(EFBB)
X(TAMPERDET)
X(TAMPERTYPE0)
X(TAMPERTYPE1)
X(TAMPERSRC0)
X(TAMPERSRC1)
X(TAMPERDETEN)
X(TAMPERLOCKSRC)
X(TAMPERDETCLK)

X(RESEN)
X(PULLUPEN)
X(SLEWRATE)

X(TRELLIS_COMB_CARRY0)
X(TRELLIS_COMB_CARRY1)
