X(MISTRAL_COMB)
X(MISTRAL_MCOMB)

X(MISTRAL_FF)
X(LAB)
X(MLAB)

X(MISTRAL_IO)
X(I)
X(O)
X(OE)
X(PAD)
X(MISTRAL_IB)
X(MISTRAL_OB)

X(A)
X(B)
X(C)
X(D)
X(E)
X(F)
X(E0)
X(F0)
X(E1)
X(F1)
X(DATAIN)
X(FFT0)
X(FFT1)
X(FFB0)
X(FFB1)
X(FFT1L)
X(FFB1L)
X(CLKIN)
X(ACLR)
X(CLK)
X(ENA)
X(SCLR)
X(SLOAD)
X(SDATA)
X(Q)

X(COMBOUT)
X(SUM_OUT)
X(CI)
X(SHAREIN)
X(CO)
X(SHAREOUT)

X(CARRY_START)
X(SHARE_START)
X(CARRY_END)
X(SHARE_END)

X(MISTRAL_ALUT6)
X(MISTRAL_ALUT5)
X(MISTRAL_ALUT4)
X(MISTRAL_ALUT3)
X(MISTRAL_ALUT2)
X(MISTRAL_NOT)
X(MISTRAL_BUF)
X(MISTRAL_CONST)
X(MISTRAL_ALUT_ARITH)
X(LUT)
X(LUT0)
X(LUT1)

X(D0)
X(D1)
X(SO)

X(WIRE)

X(GND)
X(VCC)
X(Y)

X(LOC)

X(MISTRAL_CLKENA)
X(MISTRAL_CLKBUF)

X(WA0)
X(WA1)
X(WA2)
X(WA3)
X(WA4)
X(WCLK)
X(WE)

X(MISTRAL_MLAB)
X(CLK1)
X(A1EN)
X(B1EN)
X(A1DATA)
X(B1DATA)
X(WCLK_INV)
X(WE_INV)

X(cyclonev_oscillator)
X(cyclonev_hps_interface_mpu_general_purpose)

X(MISTRAL_M10K)
X(INIT)
X(ADDRSTALLA)
X(ADDRSTALLB)
X(CFG_ABITS)
X(CFG_DBITS)

X(clkout)
X(clkout1)
X(compress_rbf)
X(oscena)
X(placer)
X(router)
X(step)
