X(FABULOUS_LC)
X(FABULOUS_COMB)
X(FABULOUS_FF)

X(SET_NORESET)
X(ASYNC_SR)
X(NEG_CLK)
X(FF)
X(LATCH_NOFF)

X(IO_1_bidirectional_frame_config_pass)
X(InPass4_frame_config)
X(OutPass4_frame_config)
X(InPass4_frame_config_mux)
X(OutPass4_frame_config_mux)
X(RegFile_32x4)
X(MULADD)
X(MUX8LUT_frame_config)
X(MUX8LUT_frame_config_mux)

X(CLK)
X(I)
X(T)
X(O)
X(Q)
X(Ci)
X(Co)

X(X0Y0)

X(REG_CLK)
X(LUT_CLK)
X(global_clock)

X(Global_Clock)
X(O2Q)

X(WRITE_DATA)
X(WRITE_ADDRESS)
X(READ_DATA)
X(READ_ADDRESS)
X(DSP_DATA_OUT)
X(DSP_DATA_IN)
X(DSP_CLR)

X(carry_in)
X(carry_out)

X(LUTFF)
X(LUTFF_E)
X(LUTFF_SR)
X(LUTFF_SS)
X(LUTFF_ESR)
X(LUTFF_ESS)
X(LUTFF_R)
X(LUTFF_S)
X(LUTFF_ER)
X(LUTFF_ES)

X(LUTFF_N)
X(LUTFF_NE)
X(LUTFF_NSR)
X(LUTFF_NSS)
X(LUTFF_NESR)
X(LUTFF_NESS)
X(LUTFF_NR)
X(LUTFF_NS)
X(LUTFF_NER)
X(LUTFF_NES)

X(clr)

X(__disconnected)
X(INIT)

X(E)
X(C)
X(D)
X(S)
X(R)
X(SR)
X(EN)

X(BEL)
X(PAD)
X(LUT1)
X(LUT4_HA)

X(BelBegin)
X(BelEnd)
X(GlobalClk)
X(CFG)

X(FABULOUS_MUX2)
X(FABULOUS_MUX4)
X(FABULOUS_MUX8)

X(M_AB)
X(M_AD)
X(M_EF)
X(M_AH)

X(I0MUX)

X(GND0)
X(GND)
X(VCC0)
X(VCC)
X(VDD0)
X(VDD)

X(_CONST0_DRV)
X(_CONST1_DRV)

X(_CONST0)
X(_CONST1)

X(_LUT_PERM)
X(_LUT_PERM_IN)
X(I_reg)
