X(SLICE_LUTX)
X(SLICE_FFX)
X(F7MUX)
X(F8MUX)
X(F9MUX)
X(CARRY4)
X(CARRY8)
X(CLEL_L)
X(CLEL_R)
X(CLEM)
X(CLEM_R)
X(CLBLL_L)
X(CLBLL_R)
X(CLBLM_L)
X(CLBLM_R)

X(A1)
X(A2)
X(A3)
X(A4)
X(A5)
X(A6)
X(CLK)
X(DI1)
X(DI2)
X(SIN)
X(WA1)
X(WA2)
X(WA3)
X(WA4)
X(WA5)
X(WA6)
X(WA7)
X(WA8)
X(WA9)
X(WE)
X(MC31)
X(O5)
X(O6)

X(CE)
X(CK)
X(D)
X(SR)
X(Q)

X(AX)
X(BX)
X(CIN)
X(CX)
X(DI0)
X(DI3)
X(DI4)
X(DI5)
X(DI6)
X(DI7)
X(DX)
X(EX)
X(FX)
X(HX)
X(S0)
X(S1)
X(S2)
X(S3)
X(S4)
X(S5)
X(S6)
X(S7)
X(CO0)
X(CO1)
X(CO2)
X(CO3)
X(CO4)
X(CO5)
X(CO6)
X(CO7)
X(O0)
X(O1)
X(O2)
X(O3)
X(O4)
X(O7)

X(OUT)

X(PSEUDO_GND)
X(PSEUDO_VCC)
X(HARD0)
X(Y)

X(BRAM)
X(RAMBFIFO36)
X(RAMBFIFO18)

X(RAMBFIFO18E2_RAMBFIFO18E2)
X(RAMB18E2_RAMB18E2)
X(FIFO18E2_FIFO18E2)

X(RAMBFIFO36E2_RAMBFIFO36E2)
X(RAMB36E2_RAMB36E2)
X(FIFO36E2_FIFO36E2)

X(BRAM_L)
X(BRAM_R)

X(RAMB18E1_RAMB18E1)
X(FIFO18E1_FIFO18E1)

X(RAMBFIFO36E1_RAMBFIFO36E1)
X(RAMB36E1_RAMB36E1)
X(FIFO36E1_FIFO36E1)

X(URAM288)
X(BEL_URAM288)
X(URAM_URAM_FT)
X(URAM_URAM_DELAY_FT)

X(DSP)
X(DSP48E2)
X(DSP_PREADD)
X(DSP_PREADD_DATA)
X(DSP_A_B_DATA)
X(DSP_MULTIPLIER)
X(DSP_M_DATA)
X(DSP_ALU)
X(DSP_OUTPUT)
X(DSP_C_DATA)

X(DSP_L)
X(DSP_R)
X(DSP48E1)
X(DSP48E1_DSP48E1)

X(HPIO_L)
X(XIPHY_BYTE_L)

X(BITSLICE_COMPONENT_RX_TX)
X(OUT_FF)
X(ODELAYE3)
X(OSERDESE3)
X(IDELAYE3)
X(ISERDESE3)
X(IN_FF)

X(LIOB33)
X(RIOB33)

X(IOB33M)
X(IOB33S)
X(IOB33)
X(IPAD)

X(PAD)
X(IOB33M_OUTBUF)
X(IOB33S_OUTBUF)
X(IOB33_OUTBUF)
X(IOB33M_INBUF_EN)
X(IOB33S_INBUF_EN)
X(IOB33_INBUF_EN)
X(IOB33M_TERM_OVERRIDE)
X(IOB33S_TERM_OVERRIDE)
X(IOB33_TERM_OVERRIDE)
X(PULL_OR_KEEP1)

X(IDELAYE2)
X(IDELAYE2_IDELAYE2)
X(OLOGICE3)
X(OLOGICE3_TFF)
X(OLOGICE3_OUTFF)
X(OLOGICE3_MISR)
X(OSERDESE2)
X(OSERDESE2_OSERDESE2)

X(ILOGICE3)
X(ILOGICE3_IFF)
X(ILOGICE3_ZHOLD_DELAY)
X(ISERDESE2)
X(ISERDESE2_ISERDESE2)

X(BUFR_BUFR)
X(BUFIO_BUFIO)
X(IDELAYCTRL_IDELAYCTRL)

X(PLL_SELECT_BEL)

X(PLL)
X(PLL_PLL_TOP)
X(MMCM)
X(MMCM_MMCM_TOP)
X(BUFGCE)
X(BUFGCE_DIV)
X(BUFCE_BUFG_PS)
X(CMT_L)

X(INTENT_DEFAULT)
X(NODE_OUTPUT)
X(NODE_DEDICATED)
X(NODE_GLOBAL_VDISTR)
X(NODE_GLOBAL_HROUTE)
X(NODE_GLOBAL_HDISTR)
X(NODE_PINFEED)
X(NODE_PINBOUNCE)
X(NODE_LOCAL)
X(NODE_HLONG)
X(NODE_SINGLE)
X(NODE_DOUBLE)
X(NODE_HQUAD)
X(NODE_VLONG)
X(NODE_VQUAD)
X(NODE_OPTDELAY)
X(NODE_GLOBAL_VROUTE)
X(NODE_GLOBAL_LEAF)
X(NODE_GLOBAL_BUFG)

X(NODE_LAGUNA_DATA)
X(NODE_CLE_OUTPUT)
X(NODE_INT_INTERFACE)
X(NODE_LAGUNA_OUTPUT)

X(BENTQUAD)
X(BOUNCEACROSS)
X(BOUNCEIN)
X(BUFGROUT)
X(BUFINP2OUT)
X(CLKPIN)
X(DOUBLE)
X(GENERIC)
X(GLOBAL)
X(HLONG)
X(HQUAD)
X(HVCCGNDOUT)
X(INPUT)
X(IOBIN2OUT)
X(IOBINPUT)
X(IOBOUTPUT)
X(LUTINPUT)
X(OPTDELAY)
X(OUTBOUND)
X(OUTPUT)
X(PADINPUT)
X(PADOUTPUT)
X(PINBOUNCE)
X(PINFEED)
X(PINFEEDR)
X(REFCLK)
X(SINGLE)
X(SLOWSINGLE)
X(SVLONG)
X(VLONG)
X(VLONG12)
X(VQUAD)

X(INTENT_SITE_WIRE)
X(INTENT_SITE_GND)

X(IOB_IBUFCTRL)
X(IOB_INBUF)
X(IOB_OUTBUF)
X(IOB_PAD)
X(TRIBUF)

X(BUFGCTRL)
X(BUFGCE_DIV_BUFGCE_DIV)
X(BUFCE_BUFCE)

X(PS7)
X(PS7_PS7)

X(0)
X(1)
X(A)
X(A0)
X(ADDRARDADDRL15)
X(ADDRATIEHIGH0)
X(ADDRATIEHIGH1)
X(ADDRBTIEHIGH0)
X(ADDRBTIEHIGH1)
X(ADDRBWRADDRL15)
X(AREG)
X(BEL)
X(BITSLICE_CONTROL_BEL)
X(BREG)
X(BUFG)
X(BUFG_BUFG)
X(BUFG_PS)
X(BUFHCE)
X(BUFMRCE)
X(C)
X(CARRYIN)
X(CARRY_TYPE)
X(CB)
X(CDDCREQ)
X(CE0)
X(CE1)
X(CEA1)
X(CEA2)
X(CEAD)
X(CEALUMODE)
X(CEB1)
X(CEB2)
X(CEC)
X(CECARRYIN)
X(CECTRL)
X(CED)
X(CEINMODE)
X(CEM)
X(CEP)
X(CFGLUT5)
X(CI)
X(CINVCTRL)
X(CINVCTRL_SEL)
X(CKB)
X(CK_C)
X(CLKARDCLK)
X(CLKB)
X(CLKBWRCLK)
X(CLKDIV)
X(CLKDIVP)
X(CLKFBIN)
X(CLKIN)
X(CLKIN1)
X(CLKIN2)
X(CLKINSEL)
X(CLKRSVD0)
X(CLKRSVD1)
X(CLK_B)
X(CLK_EXT)
X(CLR)
X(COMPENSATION)
X(CONVSTCLK)
X(CPLLLOCKDETCLK)
X(CYINIT)
X(D1)
X(D2)
X(D3)
X(D4)
X(D5)
X(D6)
X(D7)
X(D8)
X(DATAOUT)
X(DATA_RATE)
X(DATA_RATE_OQ)
X(DATA_RATE_TQ)
X(DATA_WIDTH)
X(DCITERMDISABLE)
X(DCLK)
X(DDLY)
X(DDR_CLK_EDGE)
X(DELAY_SRC)
X(DEN)
X(DI)
X(DIADI0)
X(DIADI1)
X(DIBDI0)
X(DIBDI1)
X(DIFFINBUF)
X(DIFFI_IN)
X(DIFF_IN_N)
X(DIFF_IN_P)
X(DIPADIP0)
X(DIPADIP1)
X(DIPBDIP0)
X(DIPBDIP1)
X(DLY_TEST_IN)
X(DMONITORCLK)
X(DOA_REG)
X(DOB_REG)
X(DPO)
X(DRIVE)
X(DRPCLK)
X(DWE)
X(E)
X(ECCPIPECE)
X(ECCPIPECEL)
X(ENARDEN)
X(ENBWREN)
X(EN_A)
X(EN_B)
X(EN_VTC)
X(FDCE)
X(FDCE_1)
X(FDPE)
X(FDPE_1)
X(FDRE)
X(FDRE_1)
X(FDSE)
X(FDSE_1)
X(FIFO18E1)
X(FIFO18E2)
X(FIFO36E1)
X(FIFO36E2)
X(G)
X(GND)
X(GTGREFCLK)
X(GTGREFCLK0)
X(GTGREFCLK1)
X(GTHE2_CHANNEL)
X(GTHE2_COMMON)
X(GTPE2_CHANNEL)
X(GTPE2_COMMON)
X(GTXE2_CHANNEL)
X(GTXE2_COMMON)
X(HARD_SYNC)
X(HCLK_IOI)
X(HCLK_IOI3)
X(HIGH_PERFORMANCE_MODE)
X(HPIO_OUTINV)
X(HPIO_RIGHT)
X(HPIO_VREF)
X(I)
X(I0)
X(I1)
X(I2)
X(I3)
X(I4)
X(I5)
X(IB)
X(IBUF)
X(IBUFCTRL)
X(IBUFDISABLE)
X(IBUFDS)
X(IBUFDSE3)
X(IBUFDS_DIFF_OUT)
X(IBUFDS_DIFF_OUT_IBUFDISABLE)
X(IBUFDS_DIFF_OUT_INTERMDISABLE)
X(IBUFDS_GTE3)
X(IBUFDS_GTE4)
X(IBUFDS_INTERMDISABLE)
X(IBUFDS_INTERMDISABLE_INT)
X(IBUFE3)
X(IBUF_ANALOG)
X(IBUF_IBUFDISABLE)
X(IBUF_INTERMDISABLE)
X(IDATAIN)
X(IDDR)
X(IDDRE1)
X(IDDR_2CLK)
X(IDELAYCTRL)
X(IDELAY_TYPE)
X(IDELAY_VALUE)
X(IFD_CE)
X(IGNORE0)
X(IGNORE1)
X(IN)
X(INBUF)
X(INIT)
X(INIT_OQ)
X(INIT_OUT)
X(INIT_Q1)
X(INIT_Q2)
X(INIT_TQ)
X(INJECTDBITERR)
X(INJECTSBITERR)
X(INTENT)
X(INTERFACE_TYPE)
X(INTERMDISABLE)
X(INV)
X(INVERTER)
X(IN_TERM)
X(IO)
X(IOB)
X(IOB18M_OUTBUF_DCIEN)
X(IOB18_INBUF_DCIEN)
X(IOB18_OUTBUF_DCIEN)
X(IOBDELAY)
X(IOBUF)
X(IOBUFDS)
X(IOBUFDSE3)
X(IOBUFDS_DCIEN)
X(IOBUFDS_DIFF_OUT)
X(IOBUFDS_DIFF_OUT_DCIEN)
X(IOBUFDS_DIFF_OUT_INTERMDISABLE)
X(IOBUFE3)
X(IOBUF_DCIEN)
X(IOBUF_INTERMDISABLE)
X(IOB_DIFFINBUF)
X(IOB_DIFFI_IN0)
X(IOB_O0)
X(IOB_O_IN1)
X(IOB_O_OUT0)
X(IOB_PADOUT1)
X(IOB_T0)
X(IOB_T_IN1)
X(IOB_T_OUT0)
X(IOL_IDDR)
X(IOL_OPTFF)
X(IOSTANDARD)
X(IS_CARRYIN_INVERTED)
X(IS_CE0_INVERTED)
X(IS_CE1_INVERTED)
X(IS_CLKINSEL_INVERTED)
X(IS_CLK_INVERTED)
X(IS_CLR_INVERTED)
X(IS_C_INVERTED)
X(IS_DATAIN_INVERTED)
X(IS_D_INVERTED)
X(IS_IDATAIN_INVERTED)
X(IS_IGNORE0_INVERTED)
X(IS_IGNORE1_INVERTED)
X(IS_OCLK_INVERTED)
X(IS_ODATAIN_INVERTED)
X(IS_PRE_INVERTED)
X(IS_PWRDWN_INVERTED)
X(IS_RST_INVERTED)
X(IS_R_INVERTED)
X(IS_S0_INVERTED)
X(IS_S1_INVERTED)
X(IS_SR_INVERTED)
X(IS_S_INVERTED)
X(IS_WCLK_INVERTED)
X(LDCE)
X(LDPE)
X(LDPIPEEN)
X(LI)
X(LOAD)
X(LOC)
X(LUT1)
X(LUT2)
X(LUT3)
X(LUT4)
X(LUT5)
X(LUT6)
X(LUT6_2)
X(LUT_OR_MEM5LRAM)
X(LUT_OR_MEM6LRAM)
X(MMCME2_ADV)
X(MMCME2_ADV_MMCME2_ADV)
X(MMCME2_BASE)
X(MMCME3_ADV)
X(MMCME3_BASE)
X(MMCME4_ADV)
X(MMCME4_BASIC)
X(MUXCY)
X(MUXF7)
X(MUXF8)
X(MUXF9)
X(MUX_TREE_ROOT)
X(NUM_CE)
X(O)
X(OB)
X(OBUF)
X(OBUFDS)
X(OBUFDS_GTE3)
X(OBUFDS_GTE3_ADV)
X(OBUFDS_GTE4)
X(OBUFDS_GTE4_ADV)
X(OBUFT)
X(OBUFTDS)
X(OBUFT_DCIEN)
X(OCE)
X(OCLK)
X(OCLKB)
X(ODATAIN)
X(ODDR)
X(ODDRE1)
X(ODDR_MODE)
X(ODELAYE2)
X(ODELAYE2_ODELAYE2)
X(ODELAY_TYPE)
X(ODELAY_VALUE)
X(OFB)
X(OFD_CE)
X(OLOGICE2_TFF)
X(OLOGICE2_OUTFF)
X(OQ)
X(OR2L)
X(OSC_EN)
X(OSERDES_T_BYPASS)
X(O_B)
X(PACKAGE_PIN)
X(PHASER_IN)
X(PHASER_IN_PHY)
X(PHASER_OUT)
X(PHASER_OUT_PHY)
X(PHASER_REF)
X(PIP)
X(PIPE_SEL)
X(PLL0LOCKDETCLK)
X(PLL1LOCKDETCLK)
X(PLLE2_ADV)
X(PLLE2_ADV_PLLE2_ADV)
X(PLLE2_BASE)
X(PLLE3_ADV)
X(PLLE3_BASE)
X(PLLE4_ADV)
X(PLLE4_BASIC)
X(PRE)
X(PS8)
X(PSCLK)
X(PSEN)
X(PSEUDO_GND_WIRE_GLBL)
X(PSEUDO_GND_WIRE_ROW)
X(PSEUDO_VCC_WIRE_GLBL)
X(PSEUDO_VCC_WIRE_ROW)
X(PSINCDEC)
X(PSS_ALTO_CORE)
X(PULLTYPE)
X(PWRDWN)
X(Q0)
X(Q1)
X(Q2)
X(QPLLLOCKDETCLK)
X(R)
X(RADR0)
X(RADR1)
X(RADR2)
X(RADR3)
X(RADR4)
X(RADR5)
X(RAM128X1D)
X(RAM128X1S)
X(RAM256X1D)
X(RAM256X1S)
X(RAM32M)
X(RAM32M16)
X(RAM32X1D)
X(RAM32X1S)
X(RAM32X2S)
X(RAM512X1D)
X(RAM512X1S)
X(RAM64M)
X(RAM64M8)
X(RAM64X1D)
X(RAM64X1S)
X(RAM64X8SW)
X(RAMB18E1)
X(RAMB18E2)
X(RAMB36E1)
X(RAMB36E2)
X(RAMD32)
X(RAMD64E)
X(RDB_WR_A)
X(RDB_WR_B)
X(RDCLK)
X(RDEN)
X(RDY)
X(REGCEAREGCE)
X(REGCEB)
X(REGRST)
X(RIU_NIBBLE_SEL)
X(RST)
X(RSTA)
X(RSTALLCARRYIN)
X(RSTALUMODE)
X(RSTB)
X(RSTC)
X(RSTCTRL)
X(RSTD)
X(RSTINMODE)
X(RSTM)
X(RSTP)
X(RSTRAMARSTRAM)
X(RSTRAMB)
X(RSTREG)
X(RSTREGARSTREG)
X(RSTREGB)
X(RST_A)
X(RST_B)
X(RST_DLY)
X(RST_DLY_EXT)
X(RXTX_BITSLICE)
X(RXUSRCLK)
X(RXUSRCLK2)
X(RX_BITSLICE)
X(RX_CLK)
X(RX_RST)
X(RX_RST_DLY)
X(S)
X(SELMUX2_1)
X(SERDES_MODE)
X(SIGVALIDCLK)
X(SLEEP)
X(SLEW)
X(SPO)
X(SRI)
X(SRL16E)
X(SRLC32E)
X(SRTYPE)
X(SRVAL_OQ)
X(SRVAL_TQ)
X(SYSMONE1)
X(SYSMONE4)
X(T)
X(T1)
X(T2)
X(T3)
X(T4)
X(TBYTE_IN0)
X(TBYTE_IN1)
X(TBYTE_IN2)
X(TBYTE_IN3)
X(TCE)
X(TQ)
X(TRI)
X(TXPHDLYTSTCLK)
X(TXUSRCLK)
X(TXUSRCLK2)
X(TX_BITSLICE)
X(TX_BITSLICE_TRI)
X(TX_BIT_CTRL_OUT0)
X(TX_CLK)
X(TX_RST)
X(TX_RST_DLY)
X(T_OUT)
X(URAM288_BASE)
X(USE_DPORT)
X(VCC)
X(VREF)
X(VTC_RDY)
X(WCLK)
X(WEA0)
X(WEA1)
X(WEA2)
X(WEA3)
X(WRCLK)
X(WREN)
X(WRITE_WIDTH_A)
X(WRITE_WIDTH_B)
X(XADC)
X(XORCY)
X(X_FFSYNC)
X(X_FF_AS_LATCH)
X(X_IOB_SITE_TYPE)
X(X_IO_BEL)
X(X_IO_DIR)
X(X_LUT_AS_DRAM)
X(X_LUT_AS_SRL)
X(X_ORIG_MACRO_PRIM)
X(X_ORIG_PORT_DIADI1)
X(X_ORIG_PORT_DIBDI1)
X(X_ORIG_PORT_O5)
X(X_ORIG_PORT_O6)
X(X_ORIG_PORT_SR)
X(X_ORIG_TYPE)
X(placer)
X(route)
X(router)
X(step)
X(xilinx)
