.wy-table-responsive table td,
.wy-table-responsive table th {
    white-space: initial !important;
}
.rst-content table.docutils td {
    vertical-align: top !important;
}
div[class^='highlight'] pre {
    white-space: pre;
    white-space: pre-wrap;
}
