/*
 *  nextpnr -- Next Generation Place and Route
 *
 *  Copyright (C) 2021  gatecat <gatecat@ds0.me>
 *
 *  Permission to use, copy, modify, and/or distribute this software for any
 *  purpose with or without fee is hereby granted, provided that the above
 *  copyright notice and this permission notice appear in all copies.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 *  WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 *  ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 *  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 *  ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 *  OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef BASE_CLUSTERINFO_H
#define BASE_CLUSTERINFO_H

#include "idstring.h"
#include "nextpnr_namespaces.h"

#include <vector>

NEXTPNR_NAMESPACE_BEGIN

struct CellInfo;

// The 'legacy' cluster data, used for existing arches and to provide a basic implementation for arches without complex
// clustering requirements
struct BaseClusterInfo
{
    std::vector<CellInfo *> constr_children;
    int constr_x = 0;          // this.x - parent.x
    int constr_y = 0;          // this.y - parent.y
    int constr_z = 0;          // this.z - parent.z
    bool constr_abs_z = false; // parent.z := 0
};

NEXTPNR_NAMESPACE_END

#endif /* BASE_ARCH_H */
